c (c) R T Garrod, 2021-. All rights reserved.
c ======================================================================
c  _/_/_/_/_/_/_/          _/            _/
c       _/                              _/
c      _/      _/_/_/    _/   _/_/_/   _/  _/   _/_/_/
c     _/      _/    _/  _/  _/        _/  _/  _/
c    _/      _/        _/  _/        _/_/      _/_/_/
c   _/      _/        _/  _/        _/  _/         _/
c  _/      _/        _/    _/_/_/  _/    _/  _/_/_/
c
c *Trainer for Interstellar Chemical Kinetics Simulations*
c ======================================================================
c  ...a primarily gas-phase astrochemical model for training/educational 
c     purposes, based on MAGICKAL:
c  *Model for Astrophysical Gas and Ice Chemical Kinetics And Layering*
c ======================================================================
c
c Requires files:
c   'trx_header.f'    -- dimensions and common blocks
c   'trx_io.f'        -- most input/output routines
c   'trx_shield.f'    -- H2 and CO self-shielding subroutines
c   'dgear.f'         -- integrator package
c   'trx_control.d'   -- main user interface
c   'trx_reactions.d' -- species and reactions
c   'trx_ssdata.d'    -- H2 and CO self-shielding data (must be present)
c
c Produces files:
c   'top_output.d'    -- fractional abundances, physical conditions, etc
c   'top_order.d'     -- order file containing most important reactions
c   'top_intinfo.d'   -- integrator info
c   'top_rates.d'     -- grain surface reaction rates
c
c ======================================================================
c *** MAIN PROGRAM *****************************************************
c ======================================================================

      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f'

      INTEGER IWORK(NSMAX+22)
      INTEGER NXSTEP,LEN,LENIW
      REAL*8 WORK((NSMAX+10)*NSMAX+204),Y(NSMAX)

      VERS='v2.3 (2022)'

c ----------------------------------------------------------------------

      CALL FILESET
      CALL WRITENAME(NERR)
      CALL READINPUT(EPS,HMAX,NXSTEP)
      CALL INITIAL

      IT=1
      CALL START(TOUT,NSTATE,NTASK,IROOT,IERROR,MINT,MITER,
     1           IMPL,ML,MU,MXORD,LENW,LENIW)
      CALL WRITEHEAD(EPS,HMAX,NXSTEP)

      CALL UPDATE

      DO IT=2,NOPMAX
         TIME=TIMS(IT)
         WRITE(NERR,5) 'IT=',IT,', TIME=',TIME/TYEAR,' yrs'
    5    FORMAT (A3,I5,A7,1PE13.6,A4)
         CALL EVOLVE(T,Y,NSTATE,TOUT,NTASK,IROOT,EPS,IERROR,
     1               MINT,MITER,IMPL,ML,MU,MXORD,HMAX,WORK,
     2               LENW,IWORK,LENIW,NXSTEP)
         CALL UPDATE
         IF (IT.EQ.NT) GO TO 10
      ENDDO
   10 CONTINUE
 
      CALL OUTPUT

      WRITE(NERR,*) '  DONE'

      STOP 
      END 

c ======================================================================
c ======================================================================
      SUBROUTINE FILESET
      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f'

      RFNAME='trx_reactions_deut_v2_re.d'

      NERR=6
      NJR=7
      NCON=8
      NSSH=9
      NOUT=10
      NORD=11
      NRAT=12
      NINF=13
      NCSV=14
      NLOG=15

      OPEN (UNIT=NJR,FILE=RFNAME,STATUS='OLD')

      OPEN (UNIT=NCON,FILE='trx_control.d',STATUS='OLD')
      OPEN (UNIT=NSSH,FILE='trx_ssdata.d',STATUS='OLD')

      OPEN (UNIT=NOUT,FILE='top_output.d',STATUS='UNKNOWN')
      OPEN (UNIT=NCSV,FILE='top_output.csv',STATUS='UNKNOWN')
      OPEN (UNIT=NLOG,FILE='top_log-log.csv',STATUS='UNKNOWN')

      OPEN (UNIT=NORD,FILE='top_order.d',STATUS='UNKNOWN')
      OPEN (UNIT=NRAT,FILE='top_rates.d',STATUS='UNKNOWN')
      OPEN (UNIT=NINF,FILE='top_intinfo.d',STATUS='UNKNOWN')


      RETURN 
      END

c ======================================================================
c ======================================================================
      SUBROUTINE INITIAL
      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f'

      REAL*8 MASS(NEMAX)
      CHARACTER*72 REQNAME

c --- Assign elemental masses
      DO K=1,NEMAX
         IF (ELNAME(K).EQ.'H       ') MASS(K)=1
         IF (ELNAME(K).EQ.'D       ') MASS(K)=2
         IF (ELNAME(K).EQ.'He      ') MASS(K)=4
         IF (ELNAME(K).EQ.'C       ') MASS(K)=12
         IF (ELNAME(K).EQ.'N       ') MASS(K)=14
         IF (ELNAME(K).EQ.'O       ') MASS(K)=16
         IF (ELNAME(K).EQ.'Na      ') MASS(K)=23
         IF (ELNAME(K).EQ.'Mg      ') MASS(K)=24
         IF (ELNAME(K).EQ.'Si      ') MASS(K)=28
         IF (ELNAME(K).EQ.'P       ') MASS(K)=31
         IF (ELNAME(K).EQ.'S       ') MASS(K)=32
         IF (ELNAME(K).EQ.'Cl      ') MASS(K)=35
         IF (ELNAME(K).EQ.'Fe      ') MASS(K)=56
      ENDDO

c --- Check for electron, H2, CO, etc species numbers
      NSE=0
      NSH2=0
      NSCO=0
      NSJH=0
      DO I=1,NSMAX
         IF (SPEC(I).EQ.'e-      ') NSE=I
         IF (SPEC(I).EQ.'H2      ') NSH2=I
         IF (SPEC(I).EQ.'CO      ') NSCO=I
         IF (SPEC(I).EQ.'JH      ') NSJH=I
      ENDDO

c --- Calculate masses
      DO I=1,NSMAX 
         MSUM=0
         DO K=1,NEMAX 
            MSUM=MSUM+MASS(K)*IELM(K,I) 
         ENDDO 
         AWT(I)=MSUM
         IF (SPEC(I).EQ.'e-      ') AWT(I)=1.D+0/1836.D+0
      ENDDO

c --- Find ITYPE first and last reactions
      DO I=0,NITYPE-1
         IRXSTA(I)=0
         IRXFIN(I)=0
         DO J=1,NKMAX
            IF ((ITYPE(J).EQ.I).AND.(IRXSTA(I).EQ.0)) IRXSTA(I)=J
            IF (ITYPE(J).EQ.I) IRXFIN(I)=J
         ENDDO
      ENDDO
      NREAC=0
      DO I=0,NITYPE-1
         IF (IRXFIN(I).NE.0) NREAC=NREAC+INUM(IRXFIN(I))
      ENDDO
      IF (NREAC.NE.NKMAX) THEN 
         WRITE(NERR,*)'***INUMS do not sum to NKMAX***'
         WRITE(NERR,*)
     1'**NB: Ensure full re-compilation after changing [trx_header.f]**'
         STOP 
      ENDIF

c --- Get reaction species numbers
      DO J=1,NKMAX
         DO K=1,8
            JS(K,J)=0
         ENDDO
         DO I=1,NSMAX
            DO K=1,8
               IF (RSPEC(K,J).EQ.SPEC(I)) JS(K,J)=I
            ENDDO
         ENDDO
      ENDDO

c --- Find H2 and CO photodissociation reactions
      NRH2=0
      NRCO=0
      DO J=1,NKMAX
         IF (ITYPE(J).EQ.13) THEN
            IF ((RSPEC(1,J).EQ.'H2      ').AND.
     1          (RSPEC(4,J).EQ.'H       ').AND.
     2          (RSPEC(5,J).EQ.'H       ')) NRH2=J

            IF (RSPEC(1,J).EQ.'CO      ') THEN
               IF (((RSPEC(4,J).EQ.'C       ').AND.
     1              (RSPEC(5,J).EQ.'O       ')).OR.
     2             ((RSPEC(4,J).EQ.'O       ').AND.
     3              (RSPEC(5,J).EQ.'C       '))) NRCO=J
            ENDIF
         ENDIF
      ENDDO

c --- Find surface H+H->H2 reaction
      NRJH2=0
      NRMH2=0
      DO J=1,NKMAX
         IF (ITYPE(J).EQ.14) THEN
            IF ((RSPEC(1,J).EQ.'JH      ').AND.
     1          (RSPEC(2,J).EQ.'JH      ').AND.
     2          (RSPEC(4,J).EQ.'H2     ')) NRJH2=J
         ENDIF
      ENDDO

c --- Initialise integrator output array
      DO I=1,4
         IEQP(I)=0
      ENDDO

c --- Identify reactions and multipliers for each species
      DO I=1,NSMAX
         NRX(I)=0
         DO J=1,NKMAX
            IRM=0
            DO K=1,3
               IF (JS(K,J).EQ.I) IRM=IRM-1
            ENDDO
            DO K=4,8
               IF (JS(K,J).EQ.I) IRM=IRM+1
            ENDDO
            IF (IRM.NE.0) THEN
               NRX(I)=NRX(I)+1
               IRX(NRX(I),I)=J
               RMUL(NRX(I),I)=IRM
            ENDIF
         ENDDO
      ENDDO

c --- Initialise reaction rate coefficients, etc
      CALL GRAINRATE
      IF (ISELF.NE.0) CALL INITSS

c --- Initialise reaction rates
      DO J=1,NKMAX
         XK(J)=0.D0
         R(J)=0.D0
      ENDDO

      RETURN 
      END

c ======================================================================
c ======================================================================
      SUBROUTINE START(TOUT,NSTATE,NTASK,IROOT,IERROR,MINT,MITER,
     1                 IMPL,ML,MU,MXORD,LENW,LENIW)
      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f'

      CHARACTER*8 SREAD(NSMAX)
      REAL*8 Y(NSMAX)

c Set integrator switches and constants=================================
      NTASK=ITASK
      IROOT=0
      IERROR=3
      MINT=3
      MITER=1
      IMPL=0
      ML=1
      MU=1
      MXORD=5

      NSTATE=1 
      LENIW=NSMAX+22 
      LENW=NSMAX*NSMAX+(MXORD+5)*NSMAX+2*IROOT+204

c Initialise times======================================================
      TOUT=0.0D+0
      TIME=0.0D+0
      NJAC=0

c Set initial abundances================================================
      DO I=1,NSMAX
         XN(I)=XNMIN
         DO K=1,NS0
            IF (SPEC(I).EQ.XS0(K)) THEN
               XN(I)=XN0(K)
            ENDIF
         ENDDO
      ENDDO

c Set electron abundances and atomic species via conservation===========
      DO J=1,NEMAX
         ELEMS(J)=0.0D+0
      ENDDO
      DO I=1,NSMAX 
         Y(I)=XN(I) 
         DO J=1,NEMAX
            ELEMS(J)=ELEMS(J)+IELM(J,I)*XN(I)
         ENDDO
      ENDDO

c --- Conserve electrons
      CHASUM=0.0D+0
      DO I=1,NSMAX
         IF (I.NE.NSE) CHASUM=CHASUM+ICG(I)*Y(I)
      ENDDO
      IF (CHASUM.LT.XNMIN) CHASUM=XNMIN
      Y(NSE)=CHASUM

c --- Initialize output values
      DO I=1,NSMAX 
         XN(I)=Y(I) 
      ENDDO 

c --- Set output times
      CALL TIMSET

c Initialise reaction rates=============================================
      CALL RATCON(Y)

      RETURN 
      END 

c ======================================================================
c ======================================================================
      SUBROUTINE TIMSET
      IMPLICIT REAL*16(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f'

c Set up time intervals for SDRIV3 calls================================
      NT=0
      TDEC=10.D+0**(1.0D+0/OTPD)
      TIMS(1)=0.0D+0
      DO I=2,NOPMAX
         TIMS(I)=TSTART*TDEC**REAL(I-2)
         IF (TIMS(I).GE.TFINAL*0.999D+0) THEN
           TIMS(I)=TFINAL
           IF (NT.EQ.0) NT=I
         ENDIF
      ENDDO

      IF (NT.EQ.0) THEN
         WRITE(*,*) '***NOPMAX is too small***'
         STOP
      ENDIF

      RETURN
      END

c ======================================================================
c ======================================================================
      SUBROUTINE EVOLVE (T,Y,NSTATE,TOUT,NTASK,IROOT,EPS,IERROR,
     1                   MINT,MITER,IMPL,ML,MU,MXORD,HMAX,WORK,LENW,
     2                   IWORK,LENIW,NXSTEP)
      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f'
      INTEGER IWORK(NSMAX+22),IOLD(4)
      REAL*8 Y(NSMAX),EWT(1),WORK((NSMAX+10)*(NSMAX)+204)
      EXTERNAL FCHEM,FA,G,JACOBN,USERS 

      EWT(1)=YMIN

      DO I=2,4
         IOLD(I)=IWORK(I+1)
      ENDDO

      IF (IT.LE.2) THEN
         T=0.0D+0 
         DO I=1,NSMAX 
            Y(I)=XN(I)
         ENDDO
         DO I=2,4
            IOLD(I)=0
         ENDDO
      ENDIF

      TOUT=TIME 
      CALL SDRIV3 (NSMAX,T,Y,FCHEM,NSTATE,TOUT,NTASK,IROOT,EPS,EWT,
     1             IERROR,MINT,MITER,IMPL,ML,MU,MXORD,HMAX,WORK,
     2             LENW,IWORK,LENIW,JACOBN,FA,NSMAX,NXSTEP,G,USERS)

      DO I=1,NSMAX
         XN(I)=Y(I)
      ENDDO

      IEQP(1)=IWORK(2)
      DO I=2,4
         IEQP(I)=IWORK(I+1)-IOLD(I)
      ENDDO
      SPERR=IWORK(1)

      IF (NSTATE.GT.2) STOP 

c   From 'sgear.f': 
c       The first three elements of WORK and the first five elements of
c       IWORK will contain the following statistical data:
c   WORK(1)  AVGH     The average step size used.
c   WORK(2)  HUSED    The step size last used (successfully).
c   WORK(3)  AVGORD   The average order used.
c  IWORK(1)  IMXERR   The index of the element of the solution vector that
c                     contributed most to the last error test.
c  IWORK(2)  NQUSED   The order last used (successfully).
c  IWORK(3)  NSTEP    The number of steps taken since last initialization.
c  IWORK(4)  NFE      The number of evaluations of the right hand side.
c  IWORK(5)  NJE      The number of evaluations of the Jacobian matrix.

      RETURN 
      END 

c ======================================================================
c ======================================================================
      SUBROUTINE JACOBN (N,T,Y,DFDY,MATDIM,ML,MU)
      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f'

      REAL*8 Y(N)
      REAL*8 DFDY(N,N),EPSJ,RJAC,DDIF
      REAL*8 SAVE1(N),SAVE2(N),T,YJ,YWT

      UROUND=1.11022302462516D-16
      USQRT=SQRT(UROUND)

      NJAC=0
      CALL FCHEM(N,T,Y,SAVE2)

      DO J = 1,N
         NJAC=J

         EPSJ=USQRT
         YWT=ABS(Y(J))
         RJAC = -EPSJ*YWT

         YJ = Y(J)
         Y(J) = Y(J) + RJAC
         RJAC=Y(J)-YJ

         CALL FCHEM(N,T,Y,SAVE1)
         Y(J) = YJ

         DO I = 1,N
            DFDY(I,J)=0.Q0
            DDIF=0.Q0
            IF (SAVE2(I).NE.0.Q0) DDIF=ABS((SAVE1(I)-SAVE2(I))/SAVE2(I))
            IF (DDIF.NE.0.Q0)
     1         DFDY(I,J) = (SAVE1(I) - SAVE2(I))/RJAC
         ENDDO

      ENDDO

      NJAC=0


      RETURN
      END

c ======================================================================
c ======================================================================
      SUBROUTINE FCHEM (N,T,Y,YP) 
      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f'
      REAL*8 Y(NSMAX),YP(NSMAX)

      CALL RATCON(Y)
      CALL RATADD(Y,YP)

      RETURN
      END

c ======================================================================
c ======================================================================
      SUBROUTINE RATADD(Y,YP)
      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f'

      REAL*8 Y(NSMAX),YP(NSMAX)

      DO I=1,NSMAX
         YP(I)=0.Q0
         DO J=1,NRX(I)
            YP(I)=YP(I)+RMUL(J,I)*R(IRX(J,I))
         ENDDO
      ENDDO

      RETURN
      END

c ======================================================================
c ======================================================================
      SUBROUTINE RATSIGN(X,XP1,XP2)
      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f'

      REAL*8 X(NSMAX),XP1(NSMAX),XP2(NSMAX)

      DO I=1,NSMAX
         XP1(I)=0.D0
         DO J=1,NRX(I)
            XXX=RMUL(J,I)*R(IRX(J,I))
            IF (XXX.GT.0.D0) THEN
               XP1(I)=XP1(I)+XXX
            ENDIF
         ENDDO
      ENDDO

      DO I=1,NSMAX
         XP2(I)=0.D0
         DO J=1,NRX(I)
            XXX=RMUL(J,I)*R(IRX(J,I))
            IF (XXX.LT.0.D0) THEN
               XP2(I)=XP2(I)+XXX
            ENDIF
         ENDDO
      ENDDO

      RETURN
      END

c ======================================================================
c ======================================================================
      SUBROUTINE UPDATE
      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f' 

      REAL*8 DXDTS,TERR,CERR

c Transfer results to output arrays=====================================
      DO I=1,NSMAX
         XNOP(I,IT)=XN(I)
         IF (XN(I).LT.YMIN) XNOP(I,IT)=YMIN
      ENDDO

      ISUMS=ISUMS+IEQP(2) 
      ISUMF=ISUMF+IEQP(3) 
      ISUMJ=ISUMJ+IEQP(4) 

      CALL WRITEUPDATE

c --- Write important reactions to order file
      CALL ORDER

      RETURN
      END

c ======================================================================
c ======================================================================
      SUBROUTINE RATCON(Y)
      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f'

      REAL*8 Y(NSMAX)

      T300=TEMP/300.D+0
      TI=1.D+0/TEMP
      TSQ=SQRT(TEMP)

c Gas phase ============================================================
      IF (NJAC.EQ.0) THEN
c ------ Rxn ITYPE 1
         DO J=IRXSTA(1),IRXFIN(1)
            XK(J)=A(J)*ZETA
         ENDDO

c ------ Rxn ITYPE 2-12
         DO J=IRXSTA(2),IRXFIN(12)
            XK(J)=A(J)*(T300**B(J))*DEXP(-C(J)*TI)
         ENDDO

c ------ Rxn ITYPE 13
         DO J=IRXSTA(13),IRXFIN(13)
            XK(J)=A(J)*DEXP(-C(J)*AV)*UVFAC
         ENDDO
      ENDIF

      IF (ISELF.NE.0) THEN
        IF (NJAC.EQ.0) THEN
          IF (NRH2.NE.0) CALL H2SHIELD
          IF (NRCO.NE.0) CALL COSHIELD
        ENDIF
      ENDIF

      DO J=IRXSTA(1),IRXFIN(1)
         R(J)=XK(J)*Y(JS(1,J))
      ENDDO

      DO J=IRXSTA(2),IRXFIN(12)
         R(J)=XK(J)*Y(JS(1,J))*Y(JS(2,J))*DENS
      ENDDO
      DO J=IRXSTA(13),IRXFIN(13)
         R(J)=XK(J)*Y(JS(1,J))
      ENDDO


c Grain surface ========================================================
      IF (IDUST.GE.1) THEN

c ------ ITYPE 15 - thermal desorption of atomic H
         IF (NJAC.EQ.0) THEN
            DO J=IRXSTA(15),IRXFIN(15)
               XK(J)=A(J)*DESRAT
            ENDDO
         ENDIF
         DO J=IRXSTA(15),IRXFIN(15)
            R(J)=XK(J)*Y(JS(1,J))
         ENDDO

c ------ ITYPE 99 - accretion (adsorption)
         IF (NJAC.EQ.0) THEN
            DO J=IRXSTA(99),IRXFIN(99)
               XK(J)=A(J)*CONDSP(JS(1,J))*TSQ*DENS/GD
            ENDDO
         ENDIF
         DO J=IRXSTA(99),IRXFIN(99)
            R(J)=XK(J)*Y(JS(1,J))
         ENDDO
      ENDIF

      IF (IDUST.GE.2) THEN
c ------ ITYPE 14 - JH + JH --> H2 reaction
         IF (NJAC.EQ.0) THEN
            XK(NRJH2)=A(NRJH2)*DIFRAT*GD/TNS/DENS
         ENDIF
         R(NRJH2)=XK(NRJH2)*Y(JS(1,NRJH2))*Y(JS(2,NRJH2))*DENS
      ENDIF


      RETURN
      END

c ======================================================================
c ======================================================================
      SUBROUTINE GRAINRATE
      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      INCLUDE 'trx_header.f'

c Set accretion rates===================================================
c     COND is used to calculate R_acc = (sigma_d) * <v_i> * n_i * n_d
      COND=PI*RD*RD*SQRT(8.0D+0*BOLTZ/PI/AMU)
      DO I=1,NSMAX
         CONDSP(I)=COND*STICK/SQRT(AWT(I))
      ENDDO

c Surface processes/chemistry for JH+JH->H2 reaction====================
      EDES=450.D0
      EDIF=EDES*EBFAC

c --- Set characteristic frequency for atomic H
      CHF=SQRT(2.0D+0*BOLTZ/PI/PI/AMU * SNS*EDES/AWT(NSJH))

c --- Set desorption and diffusion rates for atomic H
      DESRAT=CHF*DEXP(-EDES/DTEMP)
      DIFRAT=CHF*DEXP(-EDIF/DTEMP)

c --- Write surface reaction info to rates file
      WRITE(NRAT,*)
      WRITE(NRAT,*) 'JH + JH --> H2 reaction info:'
      WRITE(NRAT,*) '       E_des=',EDES,' K'
      WRITE(NRAT,*) '       E_dif=',EDIF,' K'
      WRITE(NRAT,*) ' E_dif:E_des=',EDIF/EDES
      WRITE(NRAT,*) '          nu=',CHF,' s-1'
      WRITE(NRAT,*) '       k_des=',DESRAT,' s-1'
      WRITE(NRAT,*) '       k_dif=',DIFRAT,' s-1'
      WRITE(NRAT,*) '      k(H+H)=',DIFRAT/TNS,' s-1'
      WRITE(NRAT,*) '------------------------------------------------'

      RETURN
      END

c ======================================================================
c ======================================================================
      SUBROUTINE DUMMY
      IMPLICIT REAL*8(A-H,O-Z),INTEGER(I-N)
      ENTRY FA (N,T,Y,A,MATDIM,ML,MU,NDE)
      RETURN
      ENTRY G (N,T,Y,NROOT)
      RETURN
      ENTRY USERS (Y,YH,YWT,SAVE1,SAVE2,T,H,EL,IMPL,N,NDE,IFLAG)
      RETURN
      ENTRY GETXK
      RETURN
      END

c ======================================================================
c ======================================================================
